import abc
import argparse
from typing import Iterable, Dict

from ..library.importance import import_module

__all__ = (
    "PluginException",
    "Plugin",
    "PluginDispatcher",)


class PluginException(BaseException):
    """Base exception."""


class Plugin(abc.ABC):
    """All apps must meet these plugin criteria."""

    @classmethod
    def find(cls, module_name: str, name: str) -> "Plugin":
        """Import all exported plugins."""

        try:
            module = import_module(f"{module_name}.shell")
        except ImportError as e:
            print(e)
            return UnavailablePlugin(name, module_name)

        for declaration in vars(module).values():
            if isinstance(declaration, type) and issubclass(declaration, Plugin) and declaration.name == name:
                return declaration()
        return UnavailablePlugin(name, module_name)

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Unique plugin name."""

    @property
    @abc.abstractmethod
    def help(self) -> str:
        """Help text associated with command."""

    @abc.abstractmethod
    def setup(self, parser: argparse.ArgumentParser):
        """Create a sub-parser for the app."""

    @abc.abstractmethod
    def main(self, parser: argparse.ArgumentParser, args: dict) -> int:
        """Run if the build app is chosen."""


class UnavailablePlugin(Plugin):
    """Fallback for imported plugins that are not available."""

    name = "unavailable"
    help = "this plugin is not available"
    module_name: str

    def __init__(self, name: str, module_name: str):
        self.name = name
        self.module_name = module_name

    def setup(self, parser: argparse.ArgumentParser):
        pass

    def main(self, parser: argparse.ArgumentParser, args: dict) -> int:
        parser.error(f"the {self.name} plugin is unavailable; please install {self.module_name}")
        return -1


class PluginDispatcher(Plugin, abc.ABC):
    """A coordinator for plugins."""

    @property
    @abc.abstractmethod
    def plugins(self) -> Iterable[Plugin]:
        """List of plugin objects."""

    _key: str
    _plugins: Dict[str, Plugin]

    def __init__(self):
        """Compute some stuff."""

        super().__init__()
        self._plugins = {plugin.name: plugin for plugin in self.plugins}
        self._key = f"{self.name}:subcommand"

    def setup(self, parser: argparse.ArgumentParser):
        """Bind all plugins."""

        subparsers = parser.add_subparsers(required=True, dest=self._key, description=self.help)
        for plugin in self._plugins.values():
            plugin.setup(subparsers.add_parser(plugin.name, help=plugin.help))

    def main(self, parser: argparse.ArgumentParser, args: dict) -> int:
        """Dispatch."""

        return self._plugins[args[self._key]].main(parser, args)



import argparse
import logging

from .plugin import Plugin, PluginDispatcher
from ..log import log


class Curricula(PluginDispatcher):
    """Aggregate all known plugins."""

    name = "command"
    help = "the subcommand corresponding to the desired module"
    plugins = (
        Plugin.find("curricula_grade", "grade"),
        Plugin.find("curricula_compile", "compile"),
        Plugin.find("curricula_format", "format"))


def main() -> int:
    """Create the parser."""

    parser = argparse.ArgumentParser(prog="curricula", description="Command line interface for Curricula")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-v", "--verbose", action="store_true", default=False)
    group.add_argument("-q", "--quiet", action="store_true", default=False)
    parser.add_argument("-l", "--log", default=None)

    curricula = Curricula()
    curricula.setup(parser)

    args = vars(parser.parse_args())
    if args["verbose"]:
        log.setLevel(logging.DEBUG)
    elif args["quiet"]:
        log.setLevel(logging.WARNING)

    if args["log"]:
        handler_stream = logging.FileHandler(args["log"])
        log.addHandler(handler_stream)

    return curricula.main(parser, args)



import importlib.util
from importlib import import_module

from pathlib import Path
from typing import Any

__all__ = (
    "import_module",
    "import_file_at_path",
    "import_module_at_path",
    "import_file_or_module_at_path")


def import_file_at_path(path: Path, module_name: str = None) -> Any:
    """Assumes the path is a file that exists."""

    if module_name is None:
        module_name = path.parts[-1].split(".", maxsplit=1)[0]

    spec = importlib.util.spec_from_file_location(module_name, str(path))
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def import_module_at_path(path: Path, module_name: str = None) -> Any:
    """Assumes that __init__.py exists in the directory."""

    if module_name is None:
        module_name = path.parts[-1]

    spec = importlib.util.spec_from_file_location(module_name, str(path.joinpath("__init__.py")))
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def import_file_or_module_at_path(path: Path, module_name: str = None) -> Any:
    """Import an object from a path."""

    if path.joinpath("__init__.py").is_file():
        return import_module_at_path(path, module_name=module_name)
    return import_file_at_path(Path(*path.parts[:-1], path.parts[-1] + ".py"), module_name=module_name)
